package libchord;

import java.math.BigInteger;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.LinkedList;

import libchord.rmi.RemoteNode;

public class Node extends UnicastRemoteObject implements RemoteNode {

	/**
	 * 
	 */
	protected static final long serialVersionUID = 127392283750209080L;
	protected boolean die;
	protected int maxSuccessors = Identifier.keySize();
	protected Identifier[] fingers = new Identifier[maxSuccessors];
	protected LinkedList<Identifier> successors = new LinkedList<Identifier>();
	protected Identifier predecessor;
	protected Identifier identifier;
	protected Thread bgThread = null;
	protected static BigInteger two = new BigInteger("2");

	public Node(Identifier identifier) throws RemoteException {
		super();
		this.identifier = identifier;
		registerMe();
		// We only know about ourselves
		successors.add(0, identifier);
		predecessor = null;
		die=false;
		BackgroundThread t = new BackgroundThread();
		bgThread = new Thread(t);
		bgThread.start();
	}

	/***
	 * Join a network
	 * 
	 * @param identifier
	 *            Our identifier
	 * @param seed
	 *            A node of the network
	 */
	public Node(Identifier identifier, Identifier seed) throws RemoteException {
		super();
		this.identifier = identifier;
		registerMe();
		join(seed);
		BackgroundThread t = new BackgroundThread();
		die =false;
		bgThread = new Thread(t);
		bgThread.start();
	}

	protected void join(Identifier seed) throws RemoteException {
		predecessor = null;
		RemoteNode remote = lookup(seed);
		successors.add(0, remote.findSuccessor(identifier.toBigInt()));
		stabilize();
	}

	protected void registerMe() throws RemoteException {
		Registry registry = LocateRegistry.getRegistry("localhost", 1099);
		registry.rebind("rmi://" + identifier.toString() + "/libchord.Node", this);
	}

	protected RemoteNode lookup(Identifier identifier) throws RemoteException {
		RemoteNode remoteNode;
		try {
			Registry registry = LocateRegistry.getRegistry("localhost", 1099);
			remoteNode = (RemoteNode) registry.lookup("rmi://" + identifier.toString() + "/libchord.Node");
		} catch (Exception e) {
			System.err.println("rmi://" + identifier.toString() + "/libchord.Node");
			log("WTF");
			// e.printStackTrace();
			throw new RemoteException(e.toString());
		}
		remoteNode.ping();
		return remoteNode;
	}

	public void leave() throws InterruptedException, RemoteException, NotBoundException {
		die = true;
		bgThread.join();
		Registry registry = LocateRegistry.getRegistry("localhost", 1099);
		registry.unbind("rmi://" + identifier.toString() + "/libchord.Node");
	}

	protected void stabilize() {
		log("Stabilizing the network");
		stabilizeSuccessors();
		try {
			RemoteNode oldSuccessor = firstAvailableSuccessor(false);
			Identifier x = oldSuccessor.getPredecessor();
			if (x != null && x.between(identifier, oldSuccessor.getIdentifier())) {
				successors.addFirst(x);
				while (successors.size() > maxSuccessors) {
					successors.removeLast();
				}
			}
			oldSuccessor.notify(identifier);
		} catch (RemoteException e) {
			System.err.println(e.getMessage());
		}
		log("Stabilizing done");
	}

	protected void stabilizeSuccessors() {
		log("Stabilizing successors");
		try {
			RemoteNode successor = firstAvailableSuccessor(false);
			LinkedList<Identifier> successorsFixed = successor.getSuccessors();
			if (!successorsFixed.isEmpty()) {
				successorsFixed.removeLast();
			}
			// Then move the first available successor to the head of the list
			successorsFixed.addFirst(successor.getIdentifier());
			// Resize the list
			while (successorsFixed.size() >= maxSuccessors) {
				successorsFixed.removeLast();
			}
			// Update my successors
			successors = successorsFixed;
		} catch (RemoteException e) {
			log("STABIL");
			log(e.getMessage());
		}
		log("Successors fixed");
	}

	protected RemoteNode firstAvailableSuccessor(boolean failedBefore) throws RemoteException {
		log("Looking for a successor");
		RemoteNode successor;
		if (successors.size() == 0) {
			log("No live successor found");
			return this;
		}
		try {
			successor = lookup(successors.getFirst());
		} catch (RemoteException re) {
			log(successors.getFirst() + " is offline");
			successors.removeFirst();
			successor = firstAvailableSuccessor(true);
		}
		if (failedBefore) {
			stabilizeSuccessors();
		}
		log("First successor is " + successors);
		return successor;
	}

	protected int fixFingers(int next) {
		log("Fixing fingers");
		// Fix fingers until i've fixed them all:
		next++;
		if (next > maxSuccessors) {
			next = 1;
		}
		try {
			fingers[next - 1] = findSuccessor(two.pow(next - 1).add(identifier.toBigInt()));
		} catch (RemoteException e) {
			System.err.println(e.getMessage());
		}
		log("Fingers Fixed");
		return next;
	}

	@Override
	public void notify(Identifier nodeID) throws RemoteException {
		log("Notify from " + nodeID);
		if ((predecessor == null) || nodeID.between(predecessor, identifier)) {
			predecessor = nodeID;
			log(nodeID + " is my predecessor");
		}
	}

	@Override
	public Identifier findSuccessor(BigInteger key) throws RemoteException {
		log("Successor of " + key + "?");
		Identifier keyID = new Identifier(key);
		RemoteNode successor = firstAvailableSuccessor(false);
		if (keyID.betweenOrRight(identifier, successor.getIdentifier())) {
			return successor.getIdentifier();
		} else {
			return lookup(closestPrecedingNode(key)).findSuccessor(key);
		}
	}

	@Override
	public Identifier closestPrecedingNode(BigInteger key) {
		Identifier keyID = new Identifier(key);
		for (int i = maxSuccessors - 1; i > 0; i--) {
			if (fingers[i] != null && fingers[i].between(identifier, keyID)) {
				return fingers[i];
			}
		}
		try {
			return firstAvailableSuccessor(false).closestPrecedingNode(key);
		} catch (Exception e) {
			return identifier;
		}
	}

	@Override
	public Identifier getIdentifier() {
		return identifier;
	}

	@Override
	public Identifier getSuccessor() throws RemoteException {
		return firstAvailableSuccessor(false).getIdentifier();
	}

	@Override
	public Identifier getPredecessor() throws RemoteException {
		return predecessor;
	}

	@Override
	public boolean ping() throws RemoteException {
		return true;
	}

	protected void log(String msg) {
		// System.out.println("[" + identifier + "]\t" + msg);
	}

	public String toString() {
		return identifier.toString();
	}

	@Override
	public LinkedList<Identifier> getSuccessors() throws RemoteException {
		return successors;
	}

	protected class BackgroundThread implements Runnable {
		public boolean die;

		public void run() {
			int next = 1;
			while (true) {
				if(die) {
					System.err.println("DEVO MORIRE");
					break;
				}
				try {
					if (predecessor != null) {
						lookup(predecessor).ping();
					}
				} catch (Exception e) {
					predecessor = null;
				}
				stabilize();
				next = fixFingers(next);
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {

				}
				if (!successors.isEmpty()) {
					 //System.out.println("[" + identifier + "] successor = " +
					 //successors.getFirst());
				}
				if (next == 1 || next == 2) {
					// System.out.println("[" + identifier + "] fingers fixed");
				}
				
			}
		}
	}
}
