package libchord;

import java.math.BigInteger;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.LinkedList;

import libchord.rmi.RemoteNode;

public class NodeNoFailure extends UnicastRemoteObject implements RemoteNode {

	/**
	 * 
	 */
	private static final long serialVersionUID = 127392283750209080L;

	// TODO: Parametrica
	protected Identifier[] fingers = new Identifier[160]; // SHA-1 size
	protected Identifier successor;
	protected Identifier predecessor;
	protected Identifier identifier;
	protected Thread bgThread = null;
	protected static BigInteger two = new BigInteger("2");

	public NodeNoFailure(Identifier identifier) throws RemoteException {
		super();
		this.identifier = identifier;
		registerMe();
		// We only know about ourselves
		successor = identifier;
		predecessor = null;
		BackgroundThread t = new BackgroundThread();
		bgThread = new Thread(t);
		bgThread.start();
	}

	/***
	 * Join a network
	 * 
	 * @param identifier
	 *            Our identifier
	 * @param seed
	 *            A node of the network
	 */
	public NodeNoFailure(Identifier identifier, Identifier seed) throws RemoteException {
		super();
		this.identifier = identifier;
		registerMe();
		join(seed);
		BackgroundThread t = new BackgroundThread();
		bgThread = new Thread(t);
		bgThread.start();
	}

	private void join(Identifier seed) throws RemoteException {
		predecessor = null;
		RemoteNode remote = lookup(seed);
		successor = remote.findSuccessor(identifier.toBigInt());
		stabilize();
	}

	protected void registerMe() throws RemoteException {
		Registry registry = LocateRegistry.getRegistry("localhost", 1099);
		registry.rebind("rmi://" + identifier.toString() + "/libchord.Node", this);
	}

	protected RemoteNode lookup(Identifier identifier) throws RemoteException {
		RemoteNode remoteNode;
		try {
			Registry registry = LocateRegistry.getRegistry("localhost", 1099);
			remoteNode = (RemoteNode) registry.lookup("rmi://" + identifier.toString() + "/libchord.Node");
		} catch (Exception e) {
			System.err.println("rmi://" + identifier.toString() + "/libchord.Node");
			log("WTF");
			// e.printStackTrace();
			throw new RemoteException(e.toString());
		}
		remoteNode.ping();
		return remoteNode;
	}

	public void leave() {
		try {
			bgThread.join(1000);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	private void stabilize() {
		log("Stabilizing the network");
		try {
			RemoteNode oldSuccessor = lookup(successor);
			Identifier x = oldSuccessor.getPredecessor();
			if (x != null && x.between(identifier, successor)) {
				successor = x;
			}
			oldSuccessor.notify(identifier);
		} catch (RemoteException e) {
			System.err.println(e.getMessage());
		}
		log("Stabilizing done");
	}

	private int fixFingers(int next) {
		log("Fixing fingers");
		next = next + 1;
		if (next > 160) {
			next = 1;
		}
		try {
			fingers[next - 1] = findSuccessor(two.pow(next - 1).add(identifier.toBigInt()));
		} catch (RemoteException e) {
			System.err.println(e.getMessage());
		}
		log("Fingers Fixed");
		return next;
	}

	@Override
	public void notify(Identifier nodeID) throws RemoteException {
		log("Notify from " + nodeID);
		if ((predecessor == null) || nodeID.between(predecessor, identifier)) {
			predecessor = nodeID;
			log(nodeID + " is my predecessor");
		}
	}

	@Override
	public Identifier findSuccessor(BigInteger key) throws RemoteException {
		log("Successor of " + key + "?");
		Identifier keyID = new Identifier(key);
		if (keyID.betweenOrRight(identifier, successor)) {
			return successor;
		} else {
			return lookup(closestPrecedingNode(key)).findSuccessor(key);
		}
	}

	@Override
	public Identifier closestPrecedingNode(BigInteger key) {
		Identifier keyID = new Identifier(key);
		for (int i = 159; i > 0; i--) {
			if (fingers[i] != null && fingers[i].between(identifier, keyID)) {
				return fingers[i];
			}
		}
		return identifier;
	}

	@Override
	public Identifier getIdentifier() {
		return identifier;
	}

	@Override
	public Identifier getSuccessor() throws RemoteException {
		return successor;
	}

	@Override
	public Identifier getPredecessor() throws RemoteException {
		return predecessor;
	}

	@Override
	public boolean ping() throws RemoteException {
		return true;
	}

	protected void log(String msg) {
		// System.out.println("[" + identifier + "]\t" + msg);
	}

	public String toString() {
		return identifier.toString();
	}

	@Override
	public LinkedList<Identifier> getSuccessors() throws RemoteException {
		return null;
	}

	protected class BackgroundThread implements Runnable {
		public void run() {
			int next = 1;
			while (true) {
				try {
					if (predecessor != null) {
						lookup(predecessor).ping();
					}
				} catch (Exception e) {
					predecessor = null;
				}
				stabilize();
				next = fixFingers(next);
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {

				}
				System.out.println("[" + identifier + "] successor = " + successor);
				if (next == 1 || next == 2) {
					System.out.println("[" + identifier + "] fingers fixed");
				}
			}
		}
	}
}
